/*! \file gsm0414.c */

#include <osmocom/gsm/protocol/gsm_04_14.h>
#include <osmocom/core/utils.h>

const struct value_string gsm414_msgt_names[] = {
	{ GSM414_MT_CLOSE_TCH_LOOP_CMD,		"CLOSE_TCH_LOOP_CMD" },
	{ GSM414_MT_CLOSE_TCH_LOOP_ACK,		"CLOSE_TCH_LOOP_ACK" },
	{ GSM414_MT_OPEN_LOOP_CMD,		"OPEN_LOOP_CMD" },
	{ GSM414_MT_CLOSE_MSLOT_LOOP_CMD,	"CLOSE_MSLOT_LOOP_CMD" },
	{ GSM414_MT_CLOSE_MSLOT_LOOP_ACK,	"CLOSE_MSLOT_LOOP_ACK" },
	{ GSM414_MT_OPEN_MSLOT_LOOP_CMD,	"OPEN_MSLOT_LOOP_CMD" },
	{ GSM414_MT_OPEN_MSLOT_LOOP_ACK,	"OPEN_MSLOT_LOOP_ACK" },
	{ GSM414_MT_ACT_EMMI_CMD,		"ACT_EMMI_CMD" },
	{ GSM414_MT_ACT_EMMI_ACK,		"ACT_EMMI_ACK" },
	{ GSM414_MT_DEACT_EMMI_CMD,		"DEACT_EMMI_CMD" },
	{ GSM414_MT_TEST_INTERFACE,		"TEST_INTERFACE" },
	{ GSM414_MT_GPRS_TEST_MODE_CMD,		"GPRS_TEST_MODE_CMD" },
	{ GSM414_MT_EGPRS_ST_RB_LOOP_CMD,	"EGPRS_START_RB_LOOP_CMD" },
	{ GSM414_MT_RESET_MS_POS_STORED,	"RESET_MS_POS_STORED" },
	{ 0, NULL }
};
